
//
// Read and discover chunks in an audio style file (IFF/SMF)
//

// A Yamaha style file is a Standard MIDI File (SMF) with additional
// proprietary chunks (e.g., OTSc, CASM). This program finds and reads
// chunks in an audio style file.

// Author:  P.J. Drongowski
// Web URL: http://sandsoftwaresound.net
// Version: 1.0
// Date:    10 April 2018
//
// Copyright (c) 2018 Paul J. Drongowski
//               Permission explicitly granted to modify and distribute

import java.lang.* ;
import java.util.* ;
import java.io.* ;

public class StyleFile {

    // Constants
    static final char[] MThd = { 'M', 'T', 'h', 'd' } ;
    static final char[] MTrk = { 'M', 'T', 'r', 'k' } ;
    static final char[] CASM = { 'C', 'A', 'S', 'M' } ;
    static final char[] AASM = { 'A', 'A', 'S', 'M' } ;
    static final char[] ASEG = { 'A', 'S', 'E', 'G' } ;
    static final char[] Adec = { 'A', 'd', 'e', 'c' } ;
    static final char[] Atab = { 'A', 't', 'a', 'b' } ;
    static final char[] AMix = { 'A', 'M', 'i', 'x' } ;
    static final char[] AFil = { 'A', 'F', 'i', 'l' } ;
    static final char[] ADSg = { 'A', 'D', 'S', 'g' } ;
    static final char[] ANdc = { 'A', 'N', 'd', 'c' } ;
    static final char[] AWav = { 'A', 'W', 'a', 'v' } ;
    static final char[] WAVE = { 'W', 'A', 'V', 'E' } ;
    static final char[] Afmt = { 'A', 'f', 'm', 't' } ;
    static final char[] Sfmt = { 'S', 'f', 'm', 't' } ;
    static final char[] Sdec = { 'S', 'd', 'e', 'c' } ;
    static final char[] Adat = { 'A', 'd', 'a', 't' } ;
    static final char[] AInf = { 'A', 'I', 'n', 'f' } ;
    static final char[] BPnt = { 'B', 'P', 'n', 't' } ;
    static final char[] OPnt = { 'O', 'P', 'n', 't' } ;
    static final char[] APnt = { 'A', 'P', 'n', 't' } ;
    static final char[] ATmp = { 'A', 'T', 'm', 'p' } ;
    static final char[] OTSc = { 'O', 'T', 'S', 'c' } ;

    // MIDI file header information
    int headerSize ;              // Should be 6 bytes
    int trackFileFormat ;         // Format 0, 1 or 2
    int numberOfTracks ;          // Should be 1 for SMF type 0
    int timingDivision ;          // Expecting 480, 1920 or some such

    // File position information
    int fileOffset = 0 ;

    // Working storage for chunk type and size information
    char[] chunkType = new char[4] ;
    int chunkSize = 0 ;
    byte dataByte = 0 ;

    // Working storage for dumping raw audio waveforms
    String audioFileName = "" ;

    private void printType(char[] type) {
	System.out.println("'" + type[0] + type[1] + type[2] + type[3] + "'") ;
    }

    // Display an array of bytes in decimal, hex and character text. Each
    // display line is eight bytes. The length of the incoming byte array
    // must be a multiple of 8.
    private void printArray(byte[] data) {
	int index = 0 ;
	for (int i = (data.length / 8) ; i > 0 ; i--) {
	    for (int j = index ; j < (index+8) ; j++) {
		System.out.printf("%4d ", data[j]) ;
	    }
	    System.out.print("| ") ;
	    for (int j = index ; j < (index+8) ; j++) {
		System.out.printf("%02X ", data[j]) ;
	    }
	    System.out.print("| ") ;
	    for (int j = index ; j < (index+8) ; j++) {
		if ((data[j] >= 32) && (data[j] < 127)) {
		    System.out.printf("%c", data[j]) ;
		} else {
		    System.out.printf(".") ;
		}
	    }
	    System.out.println("") ;
	    index = index + 8 ;
	}
    }

    // An ANdc subchunk identifies the audio file (i.e., the file name).
    // Return the subchunk size.
    private int readANdc(DataInputStream d, int offset) throws IOException {
	char[] chunkType = new char[4] ;
	int fileOffset = offset ;
	byte dataByte = 0 ;
	int subchunkSize = 0 ;

	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, ANdc)) {
	    System.out.println("*fatal* Not ANdc subchunk (no 'ANdc')") ;
	}
	subchunkSize = d.readInt() ;

	System.out.println("File Offset:       " + fileOffset) ;
	System.out.print("Chunk type:        ") ;
	printType(chunkType) ;
	System.out.println("Subchunk size:     " + subchunkSize) ;

	audioFileName = "" ;
	for (int i = 0 ; i < subchunkSize ; i++) {
	    dataByte = d.readByte() ;
	    if (dataByte != 0) {
		audioFileName = audioFileName + (char)dataByte ;
	    }
	}

	System.out.println("Audio file name:   " + audioFileName) ;

	return( subchunkSize ) ;
    }

    // An AInf chunk represents audio meta-information. It consists of
    // several subchunks: BPnt, OPnt, APnt, ATmp
    private int readAInf(DataInputStream d, int offset) throws IOException {
	char[] chunkType = new char[4] ;
	int fileOffset = offset ;
	byte dataByte = 0 ;
	int chunkSize = 0 ;

	// Read and test the AInf chunk type and chunk size
	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, AInf)) {
	    System.out.println("*fatal* Not AInf subchunk (no 'AInf')") ;
	}
	chunkSize = d.readInt() ;

	System.out.println("File Offset:       " + fileOffset) ;
	System.out.print("Chunk type:        ") ;
	printType(chunkType) ;
	System.out.println("Chunk size:        " + chunkSize) ;

	// Skip over the Ainf subchunks; probably should dump these...
	for (int i = chunkSize ; i > 0 ; i--) {
	    dataByte = d.readByte() ;
	}
	return( chunkSize ) ;
    }

    private void dumpRawWaveform(DataInputStream d, int dataSize)
	throws IOException {
	byte dataByte = 0 ;
	String fullName = audioFileName.replaceAll("\\s+","") + ".raw" ;
	System.err.println("'" + fullName + "'" ) ;
	FileOutputStream fos = new FileOutputStream(fullName) ;

	for (int i = dataSize ; i > 0 ; i--) {
	    dataByte = d.readByte() ;
	    fos.write(dataByte) ;
	}

	fos.close() ;
    }

    // An AWav chunk represents the waveform data. It consists of several
    // subchunks: WAVE, Afmt, Sfmt (including Sdec), Adat, AInf
    private int readAWav(DataInputStream d, int offset) throws IOException {
	char[] chunkType = new char[4] ;
	int fileOffset = offset ;
	byte dataByte = 0 ;
	int chunkSize = 0 ;
	int subchunkSize = 0 ;

	// Read and test the AWav chunk type and chunk size
	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, AWav)) {
	    System.out.println("*fatal* Not AWav chunk (no 'AWav')") ;
	}
	chunkSize = d.readInt() ;

	System.out.println("File Offset:       " + fileOffset) ;
	System.out.print("Chunk type:        ") ;
	printType(chunkType) ;
	System.out.println("Chunk size:        " + chunkSize) ;
	fileOffset = fileOffset + 8 ;

	// Read and test the WAVE marker
	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, WAVE)) {
	    System.out.println("*fatal* No 'WAVE' marker") ;
	}
	fileOffset = fileOffset + 4 ;

	// Read the Afmt subchunk
	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, Afmt)) {
	    System.out.println("*fatal* Not Afmt subchunk (no 'Afmt')") ;
	}
	subchunkSize = d.readInt() ;
	for (int i = subchunkSize ; i > 0 ; i--) dataByte = d.readByte() ;
	fileOffset = fileOffset + subchunkSize + 8 ;

	// Read the Sfmt and Sdec subchunks. Sfmt contains Sdec.
	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, Sfmt)) {
	    System.out.println("*fatal* Not Sfmt subchunk (no 'Sfmt')") ;
	}
	subchunkSize = d.readInt() ;
	for (int i = subchunkSize ; i > 0 ; i--) dataByte = d.readByte() ;
	fileOffset = fileOffset + subchunkSize + 8 ;

	// Read the Adat subchunk and dump raw audio waveform
	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, Adat)) {
	    System.out.println("*fatal* Not Adat subchunk (no 'Adat')") ;
	}
	subchunkSize = d.readInt() ;
	dumpRawWaveform(d, subchunkSize) ;
	fileOffset = fileOffset + subchunkSize + 8 ;

	readAInf(d, fileOffset) ;

	return( chunkSize ) ;
    }

    // Read and process the rest of an ADSg (sub)chunk. An ADSg chunk
    // represents a single audio file. ADSg contains ANdc, AWav
    private int readADSg(DataInputStream d, int offset) throws IOException {
	char[] chunkType = new char[4] ;
	int fileOffset = offset ;
	byte dataByte = 0 ;
	int chunkSize = 0 ;
	int subchunkSize = 0 ;

	// Read and test the ADSg chunk type and subchunk size
	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, ADSg)) {
	    System.out.println("*fatal* Not ADSg subchunk (no 'ADSg')") ;
	}
	chunkSize = d.readInt() ;

	System.out.println("") ;
	System.out.println("File Offset:       " + fileOffset) ;
	System.out.print("Chunk type:        ") ;
	printType(chunkType) ;
	System.out.println("Chunk size:        " + chunkSize) ;
	fileOffset = fileOffset + 8 ;

	subchunkSize = readANdc(d, fileOffset) ;
	fileOffset = fileOffset + subchunkSize + 8 ;

	subchunkSize = readAWav(d, fileOffset) ;
	fileOffset = fileOffset + subchunkSize + 8 ;

	// Return the size of the ADSg chunk
	return( chunkSize ) ;
    }

    // Read information in the AFil chunk. The AFil chunk has several nested
    // subchunks:
    //    AFil contains multiple ADSg (sub)chunks
    //    ADSg contains ANdc, AWav
    //    AWav contains WAVE, Afmt, Sfmt, Sdec, Adat
    //    AInf contains BPnt, OPnt, APnt, ATmp
    // Each ADSg represents an audio file. There is one ADSg for each of the
    // 15 possible style sections.
    private void readAFil(DataInputStream d, int offset) throws IOException {
	int fileOffset = offset ;
	char[] chunkType = new char[4] ;
	int subchunkSize = 0 ;
	byte dataByte = 0 ;

	for (int i = 15 ; i > 0 ; i--) {
	    // Read an ADSg and get its subchunk size
	    subchunkSize = readADSg(d, fileOffset) ;
	    fileOffset = fileOffset + subchunkSize + 8 ;
	}
    }

    // Read an ASEG subchunk
    private void readASEG(DataInputStream d, int offset) throws IOException {
	char[] chunkType = new char[4] ;
	int subchunkSize = 0 ;
	byte dataByte = 0 ;
	byte[] AtabData = new byte[104] ;
	byte[] AMixData = new byte[32] ;
	int index = 0 ;
	String sectionName = "" ;
	char charByte ;

	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, Adec)) {
	    System.out.println("*fatal* Not Adec subchunk (no 'Adec')") ;
	}
	subchunkSize = d.readInt() ;

	for (int i = subchunkSize ; i > 0 ; i--) {
	    charByte = (char)d.readByte() ;
	    sectionName = sectionName + charByte ;
	}
	System.out.println("Adec name:         " + sectionName) ;
	System.out.println("Subchunk size:     " + subchunkSize) ;

	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, Atab)) {
	    System.out.println("*fatal* Not Atab subchunk (no 'Atab')") ;
	}
	subchunkSize = d.readInt() ;
	System.out.print("Atab type:         ") ;
	printType(chunkType) ;
	System.out.println("Subchunk size:     " + subchunkSize) ;

	index = 0 ;
	for (int i = subchunkSize ; i > 0 ; i--) {
	    dataByte = d.readByte() ;
	    AtabData[index] = dataByte ;
	    index++ ;
	}
	printArray(AtabData) ;

	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, AMix)) {
	    System.out.println("*fatal* Not AMix subchunk (no 'AMix')") ;
	}
	subchunkSize = d.readInt() ;
	System.out.print("AMix type:         ") ;
	printType(chunkType) ;
	System.out.println("Subchunk size:     " + subchunkSize) ;

	index = 0 ;
	for (int i = subchunkSize ; i > 0 ; i--) {
	    dataByte = d.readByte() ;
	    AMixData[index] = dataByte ;
	    index++ ;
	}
	printArray(AMixData) ;
    }

    // Read an AASM chunk consisting of ASEG subchunks
    private void readAASM(DataInputStream d, int offset) throws IOException {
	int fileOffset = offset ;
	char[] chunkType = new char[4] ;
	int subchunkSize = 0 ;
	byte dataByte = 0 ;

	for (int i = 15 ; i > 0 ; i--) {
	    chunkType[0] = (char)d.readByte() ;
	    chunkType[1] = (char)d.readByte() ;
	    chunkType[2] = (char)d.readByte() ;
	    chunkType[3] = (char)d.readByte() ;
	    if (! Arrays.equals(chunkType, ASEG)) {
		System.out.println("*fatal* Not ASEG subchunk (no 'ASEG')") ;
	    }
	    subchunkSize = d.readInt() ;

	    System.out.println("") ;
	    System.out.println("File Offset:       " + fileOffset) ;
	    System.out.print("Subchunk type:     ") ;
	    printType(chunkType) ;
	    System.out.println("Subchunk size:     " + subchunkSize) ;

	    readASEG(d, fileOffset) ;

	    fileOffset = fileOffset + subchunkSize + 8 ;
	}
    }

    public void readStyleFile(DataInputStream d) throws IOException {
	fileOffset = 0 ;

	// Dump the MIDI file header just to check sanity
	chunkType[0] = (char)d.readByte() ;
	chunkType[1] = (char)d.readByte() ;
	chunkType[2] = (char)d.readByte() ;
	chunkType[3] = (char)d.readByte() ;
	if (! Arrays.equals(chunkType, MThd)) {
	    System.out.println("*fatal* Not a MIDI file (no 'MThd')") ;
	    throw new IOException("MIDI file expected") ;
	}

	headerSize = d.readInt() ;
	if (headerSize != 6) {
	    System.out.println("*warning* MIDI header size is not 6") ;
	}
	trackFileFormat = d.readShort() ;
	numberOfTracks = d.readShort() ;
	timingDivision = d.readShort() ;
	System.out.print("Header chunk type: ") ;
	printType(chunkType) ;
	System.out.println("Header chunk size: " + headerSize) ;
	System.out.println("Track file format: " + trackFileFormat) ;
	System.out.println("Number of tracks:  " + numberOfTracks) ;
	System.out.println("Timing division:   " + timingDivision) ;
	System.out.println() ;

	fileOffset = fileOffset + headerSize + 8 ;

	// Find all of the chunks and display type/size info for each
	try {
	    // Loop until the end of file exception is thrown
	    while( true ) {
		chunkType[0] = (char)d.readByte() ;
		chunkType[1] = (char)d.readByte() ;
		chunkType[2] = (char)d.readByte() ;
		chunkType[3] = (char)d.readByte() ;
		chunkSize = d.readInt() ;

		System.out.println("") ;
		System.out.println("File Offset:       " + fileOffset) ;
		System.out.print("Chunk type:        ") ;
		printType(chunkType) ;
		System.out.println("Chunk size:        " + chunkSize) ;

		if (Arrays.equals(chunkType, AASM)) {
		    readAASM(d, (fileOffset + 8)) ;
		} else if (Arrays.equals(chunkType, AFil)) {
		    readAFil(d, (fileOffset + 8)) ;
		} else {
		    for (int i = chunkSize ; i > 0 ; i--) {
			dataByte = d.readByte() ;
		    }
		}

		fileOffset = fileOffset + chunkSize + 8 ;
	    }
	} catch (EOFException e) { }

    }
}

